import asyncio
import os.path

from async_generator import async_generator, yield_, yield_from_

from PyQt5.QtCore import (pyqtSignal, QObject)

from galacteek import log
from galacteek import ensure
from galacteek.ipfs.wrappers import ipfsOp
from galacteek.ipfs.cidhelpers import joinIpfs
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.ipfsops import *  # noqa
from galacteek.ipfs import pb
from galacteek.core.asynclib import async_enterable
from galacteek.core.jtraverse import traverseParser


class DAGObj:
    def __init__(self, data):
        self.data = data


class DAGRewindException(Exception):
    pass


class DAGEditor(object):
    def __init__(self, dag):
        self.dag = dag

    @property
    def d(self):
        return self.dag.d

    async def __aenter__(self):
        log.debug('Editing DAG: {cid}'.format(cid=self.dag.dagCid))
        return self

    async def __aexit__(self, *args):
        await self.dag.sync()


class DAGOperations:
    def debug(self, msg):
        log.debug(msg)

    @ipfsOp
    async def get(self, op, path):
        self.debug('DAG get: {}'.format(os.path.join(self.dagCid, path)))
        try:
            dagNode = await op.dagGet(
                os.path.join(self.dagCid, path))
        except aioipfs.APIError as err:
            log.debug('DAG get: {0}. An error occured: {1}'.format(
                path, err.message))
            return None

        if isinstance(dagNode, dict) and 'data' in dagNode and \
                'links' in dagNode:
            """ Try and decode the protobuf data """
            msg = pb.decodeUnixfsDagNode(dagNode['data'])
            if not msg:
                return
            return msg['data']
        else:
            return dagNode

    @ipfsOp
    async def cat(self, op, path):
        return await op.client.cat(
            os.path.join(self.dagCid, path))

    @ipfsOp
    async def list(self, op, path=''):
        data = await self.get(path)
        if data is None:
            return

        if isinstance(data, dict):
            return list(data.keys())

    @ipfsOp
    async def chNode(self, ipfsop, path, start=None, create=False):
        comps = path.split('/')
        cur = start if start else self.dagRoot

        for comp in comps:
            if comp not in cur:
                if create:
                    cur[comp] = {}
                    cur = cur[comp]
                else:
                    return cur
            else:
                cur = cur[comp]

        return cur

    @ipfsOp
    async def resolve(self, op, path=''):
        return await op.dagResolve(
            os.path.join(self.dagCid, path))

    @async_generator
    async def walk(self, op, path='', maxObjSize=0, depth=1):
        """
        Do not use this!
        """

        data = await self.get(path)
        if data is None:
            return

        if isinstance(data, dict) and 'data' in data and 'links' in data:
            """ Try and decode the protobuf data """
            msg = pb.decodeUnixfsDagNode(data['data'])
            if not msg:
                return
            if maxObjSize == 0 or \
                    (maxObjSize > 0 and msg['size'] < maxObjSize):
                await yield_((path, DAGObj(msg['data'])))

        if not data:
            return

        if isinstance(data, dict):
            for objKey, objValue in data.items():
                if objKey == '/':
                    out = await op.client.cat(objValue)
                    await yield_((path, DAGObj(out)))
                if objKey == 'data' or objKey == 'links':
                    continue
                else:
                    await yield_from_(
                        self.walk(op, path=os.path.join(path, objKey),
                                  maxObjSize=maxObjSize, depth=depth)
                    )

        elif isinstance(data, list):
            for idx, obj in enumerate(data):
                await yield_from_(
                    self.walk(op, path=os.path.join(path, str(idx)),
                              maxObjSize=maxObjSize)
                )
        elif isinstance(data, str):
            await yield_((path, DAGObj(data)))


class DAGPortal(QObject, DAGOperations):
    """
    DAG portal
    """

    cidChanged = pyqtSignal(str)
    loaded = pyqtSignal(str)

    def __init__(self, dagCid=None, dagRoot=None, offline=False,
                 parent=None):
        super().__init__(parent)
        self._dagCid = dagCid
        self._dagRoot = dagRoot
        self._dagPath = IPFSPath(self._dagCid)
        self.evLoaded = asyncio.Event()
        self.offline = offline

    @property
    def d(self):
        return self._dagRoot

    def debug(self, msg):
        log.debug('DAGPortal({cid}): {msg}'.format(
            cid=self.dagCid, msg=msg))

    @property
    def dagCid(self):
        return self._dagCid

    @dagCid.setter
    def dagCid(self, val):
        self._dagCid = val
        self._dagPath = IPFSPath(self._dagCid)
        self.cidChanged.emit(val)

    @property
    def dagPath(self):
        return self._dagPath

    @property
    def dagRoot(self):
        return self._dagRoot

    def path(self, subpath):
        return os.path.join(joinIpfs(self.dagCid), subpath)

    def child(self, subpath):
        if self.dagPath.valid:
            return self.dagPath.child(subpath)

    async def waitLoaded(self, timeout=15):
        return asyncio.wait_for(self.evLoaded.wait(), timeout)

    @ipfsOp
    async def load(self, op, timeout=10):
        self._dagRoot = await op.waitFor(op.dagGet(self.dagCid), timeout)
        if self.dagRoot:
            self.evLoaded.set()
            self.loaded.emit(self.dagCid)
            return self.dagRoot

    @ipfsOp
    async def sync(self, op, timeout=10):
        try:
            cid = await op.waitFor(
                op.dagPut(self.dagRoot, offline=self.offline), timeout)
        except aioipfs.APIError:
            pass
        else:
            self.dagCid = cid

    async def __aenter__(self):
        if self.dagRoot is None:
            await self.load()
            if self.dagRoot is None:
                raise Exception('Cannot load dag')
        return self

    @async_enterable
    async def edit(self):
        return DAGEditor(self)

    async def __aexit__(self, *args):
        pass

    def link(self, cid):
        if isinstance(cid, str):
            return {"/": cid}
        elif isinstance(cid, dict) and 'Hash' in cid:
            return {"/": cid['Hash']}


class EvolvingDAG(QObject, DAGOperations):
    """
    :param str dagMetaMfsPath: the path inside the MFS for the metadata
        describing this DAG
    """

    changed = pyqtSignal()
    dagCidChanged = pyqtSignal(str)
    metadataEntryChanged = pyqtSignal()
    available = pyqtSignal(object)

    keyCidLatest = 'cidlatest'

    def __init__(self, dagMetaMfsPath, dagMetaHistoryMax=12, offline=False,
                 unpinOnUpdate=True, loop=None):
        super().__init__()

        self.lock = asyncio.Lock()
        self.loop = loop if loop else asyncio.get_event_loop()
        self.loaded = asyncio.Future()

        self._curMetaEntry = None
        self._dagRoot = None
        self._dagMeta = None
        self._dagCid = None
        self._offline = offline
        self._dagMetaMaxHistoryItems = dagMetaHistoryMax
        self._dagMetaMfsPath = dagMetaMfsPath
        self._unpinOnUpdate = unpinOnUpdate
        self.changed.connect(lambda: ensure(self.ipfsSave()))

    @property
    def dagMetaMfsPath(self):
        return self._dagMetaMfsPath  # path inside the mfs

    @property
    def dagMetaMaxHistoryItems(self):
        return self._dagMetaMaxHistoryItems

    @property
    def curMetaEntry(self):
        return self._curMetaEntry

    @curMetaEntry.setter
    def curMetaEntry(self, value):
        self._curMetaEntry = value

    @property
    def dagCid(self):
        return self._dagCid

    @dagCid.setter
    def dagCid(self, cid):
        self._dagMeta[self.keyCidLatest] = cid
        self._dagCid = cid
        self.debug("DAG's CID now is {0}".format(cid))
        self.dagCidChanged.emit(cid)

    @property
    def dagMeta(self):
        return self._dagMeta

    @property
    def dagRoot(self):
        return self._dagRoot

    @property
    def root(self):
        return self._dagRoot

    def debug(self, msg):
        log.debug('EDAG ({meta}): {msg}'.format(
            meta=self.dagMetaMfsPath, msg=msg))

    def initDag(self):
        return {}

    def updateDagSchema(self, dag):
        pass

    def mkLink(self, cid):
        if isinstance(cid, str):
            return {"/": cid}
        elif isinstance(cid, dict) and 'Hash' in cid:
            return {"/": cid['Hash']}

    async def load(self):
        await self.loadDag()

    @ipfsOp
    async def loadDag(self, op):
        self.debug('Loading DAG metadata file')
        meta = await op.filesReadJsonObject(self.dagMetaMfsPath)

        self.debug('Metadata is {}'.format(meta))
        if meta is not None:
            self._dagMeta = meta
            latest = self.dagMeta.get(self.keyCidLatest, None)
            if latest:
                self.dagCid = latest
                self.debug('Getting DAG: {cid}'.format(cid=self.dagCid))
                self._dagRoot = await op.dagGet(self.dagCid)

                if self.dagRoot:
                    if self.updateDagSchema(self.dagRoot) is True:
                        # save right away
                        await self.ipfsSave()
            else:
                self.debug('No CID history, reinitializing')
                # How inconvenient ..
                # TODO: the history could be used here to recreate a DAG
                # from previous CIDs but we really shouldn't have to enter here
                self._dagRoot = {}

            self.loaded.set_result(True)
        else:
            self._dagMeta = {
                self.keyCidLatest: None,
                'history': []
            }
            self._dagRoot = self.initDag()
            self.updateDagSchema(self.dagRoot)
            await self.ipfsSave()
            self.loaded.set_result(True)

        self.parser = traverseParser(self.dagRoot)
        self.loop.call_soon(self.available.emit, self.dagRoot)

    @ipfsOp
    async def ipfsSave(self, op):
        self.debug('Saving (acquiring lock)')
        with await self.lock:
            prevCid = self.dagCid
            history = self.dagMeta.setdefault('history', [])
            maxItems = self.dagMetaMaxHistoryItems

            # We always PIN the latest DAG and do a pin update using the
            # previous item in the history

            cid = await op.dagPut(self.dagRoot, pin=True,
                                  offline=self._offline)
            if cid is not None:
                if prevCid is not None and prevCid not in history:
                    if len(history) > maxItems:
                        # Purge old items
                        [history.pop() for idx in range(
                            0, len(history) - maxItems)]

                    history.insert(0, prevCid)
                    await op.pinUpdate(prevCid, cid, unpin=self._unpinOnUpdate)

                # Save the new CID and update the metadata
                await self.saveNewCid(cid)
            else:
                # Bummer
                self.debug('DAG could not be built')
                return False

        return True

    @ipfsOp
    async def saveNewCid(self, ipfsop, cid):
        # Save the new CID and update the metadata
        self.dagCid = cid
        await ipfsop.filesWriteJsonObject(self.dagMetaMfsPath,
                                          self.dagMeta)
        entry = await ipfsop.filesStat(self.dagMetaMfsPath)
        if entry:
            self.curMetaEntry = entry

    @ipfsOp
    async def rewind(self, ipfsop):
        """
        Using the history, rewind the EDAG in time
        (the first CID in the history becomes the new CID)

        Since we're using pinUpdate() when upgrading
        the DAG, we check that the object corresponding
        to the "previous" CID in the history is still available
        (didn't get collected) before rewriting the history.
        """

        history = self.dagMeta['history']
        prevCid = self.dagMeta[self.keyCidLatest]

        with await self.lock:
            if len(history) >= 2:
                newCid = history[0]

                # Load the DAG corresponding to the previous CID
                # Timeout is lower than the default, we can't hold
                # the lock for too long
                pDag = await ipfsop.dagGet(newCid, timeout=5)

                if not pDag:
                    # We don't have it :\
                    raise DAGRewindException(
                        'Previous object unavailable')

                # Pop it now and set the latest CID
                history.pop(0)
                self.dagMeta[self.keyCidLatest] = newCid

                # Save metadata, and save this DAG, replacing dagRoot
                # Do the pin update
                self._dagRoot = pDag
                await self.saveNewCid(newCid)

                await ipfsop.pinUpdate(prevCid, newCid,
                                       unpin=self._unpinOnUpdate)
            else:
                raise DAGRewindException('No DAG history')

    async def __aenter__(self):
        """
        Lock is acquired on entering the async ctx manager
        """
        await self.lock.acquire()
        return self

    async def __aexit__(self, *args):
        """
        Release the lock and save. The changed signal is emitted
        """
        self.lock.release()
        self.loop.call_soon(self.changed.emit)

    @async_enterable
    async def portal(self):
        return DAGPortal(dagCid=self.dagCid, dagRoot=self.dagRoot)



import os.path
import functools

from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QTreeView
from PyQt5.QtWidgets import QHeaderView
from PyQt5.QtWidgets import QToolBox
from PyQt5.QtWidgets import QProgressBar
from PyQt5.QtWidgets import QSpacerItem
from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QAbstractItemView
from PyQt5.QtWidgets import QStackedWidget

from PyQt5.QtGui import QStandardItemModel
from PyQt5.QtGui import QKeySequence

from PyQt5.QtCore import QCoreApplication
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QEvent
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QMimeData
from PyQt5.QtCore import QUrl

from galacteek.appsettings import *
from galacteek.core.modelhelpers import UneditableItem
from galacteek.ipfs import cidhelpers
from galacteek.ipfs.stat import StatInfo
from galacteek.ipfs.ipfsops import *
from galacteek.ipfs.wrappers import ipfsOp, ipfsStatOp
from galacteek.ipfs.cache import IPFSEntryCache
from galacteek.ipfs.mimetype import detectMimeType
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.cidhelpers import joinIpfs
from galacteek.ipfs.cidhelpers import cidValid
from galacteek.ipfs.cidhelpers import cidConvertBase32
from galacteek import ensure

from .i18n import *
from .helpers import *
from .widgets import GalacteekTab
from .widgets import IPFSUrlLabel
from .widgets import IPFSPathClipboardButton
from .hashmarks import *

import aioipfs


def iCIDInfo(cidv, linksCount, size):
    return QCoreApplication.translate(
        'IPFSHashExplorer',
        'CID (v{0}): {1} links, total size: {2}').format(
        cidv,
        linksCount,
        size)


def iGitClone():
    return QCoreApplication.translate('IPFSHashExplorer',
                                      'Clone repository')


def iGitClonedRepo(path):
    return QCoreApplication.translate(
        'IPFSHashExplorer',
        'Cloned git repository: {0}').format(path)


def iGitErrorCloning(msg):
    return QCoreApplication.translate(
        'IPFSHashExplorer',
        'Error cloning git repository: {0}').format(msg)


def iGitInvalid():
    return QCoreApplication.translate('IPFSHashExplorer',
                                      'Invalid git repository')


def iLoading():
    return QCoreApplication.translate('IPFSHashExplorer', 'Loading ...')


def iTimeout():
    return QCoreApplication.translate('IPFSHashExplorer',
                                      'Timeout error')


def iTimeoutTryNoResolve():
    return QCoreApplication.translate(
        'IPFSHashExplorer',
        'Timeout error: trying without resolving nodes types ..')


def iTimeoutInvalidHash():
    return QCoreApplication.translate('IPFSHashExplorer',
                                      'Timeout error: invalid hash')


class IPFSItem(UneditableItem):
    def __init__(self, text, icon=None):
        super().__init__(text, icon=icon)
        self.setParentHash(None)

    def setParentHash(self, hash):
        self.parentHash = hash

    def getParentHash(self):
        return self.parentHash


class IPFSNameItem(IPFSItem):
    def __init__(self, entry, text, icon):
        super().__init__(text, icon=icon)

        self._entry = entry
        self._mimeType = None

    @property
    def entry(self):
        return self._entry

    @property
    def ipfsPath(self):
        return IPFSPath(self.getFullPath(), autoCidConv=True)

    @property
    def mimeType(self):
        return self._mimeType

    @mimeType.setter
    def mimeType(self, mime):
        self._mimeType = mime

    def mimeFromDb(self, db):
        mType = db.mimeTypeForFile(self.entry['Name'])
        if mType:
            self.mimeType = mType.name()

    @property
    def mimeCategory(self):
        if self.mimeType:
            return self.mimeType.split('/')[0]

    def cid(self):
        return cidhelpers.getCID(cidConvertBase32(self.entry['Hash']))

    def getFullPath(self):
        """
        Returns the full IPFS path of the entry associated with this item
        (preserving file names) if we have the parent's hash, or the IPFS path
        with the entry's hash otherwise
        """
        parentHash = self.getParentHash()
        name = self.entry['Name']
        if parentHash:
            return joinIpfs(os.path.join(parentHash, name))
        else:
            return joinIpfs(cidConvertBase32(self.entry['Hash']))

    def isRaw(self):
        return self.entry['Type'] == 0

    def isDir(self):
        return self.entry['Type'] == 1

    def isFile(self):
        return self.entry['Type'] == 2

    def isMetadata(self):
        return self.entry['Type'] == 3

    def isSymlink(self):
        return self.entry['Type'] == 4

    def isUnknown(self):
        return self.entry['Type'] == -1


class IPFSHashItemModel(QStandardItemModel):
    COL_NAME = 0
    COL_SIZE = 1
    COL_MIME = 2
    COL_HASH = 3

    def __init__(self, parent):
        super(IPFSHashItemModel, self).__init__(parent)

        self.entryCache = IPFSEntryCache()
        self.rowsInserted.connect(self.onRowsInserted)

    def clearModel(self):
        self.clear()
        self.entryCache.reset()

    def mimeData(self, indexes):
        mimedata = QMimeData()

        urls = []
        for idx in indexes:
            if not idx.isValid():
                continue

            nameItem = self.getNameItemFromIdx(idx)

            if nameItem:
                url = QUrl('dweb:{}'.format(nameItem.getFullPath()))
                urls.append(url)

        mimedata.setUrls(urls)
        return mimedata

    def canDropMimeData(self, data, action, row, column, parent):
        return True

    def getHashFromIdx(self, idx):
        idxHash = self.index(idx.row(), self.COL_HASH, idx.parent())
        return self.data(idxHash)

    def getNameFromIdx(self, idx):
        idxName = self.index(idx.row(), self.COL_NAME, idx.parent())
        return self.data(idxName)

    def getNameItemFromIdx(self, idx):
        idxName = self.index(idx.row(), self.COL_NAME, idx.parent())
        return self.itemFromIndex(idxName)

    def onRowsInserted(self, parent, first, last):
        """ Update the entry cache when rows are added """
        for itNum in range(first, last + 1):
            itNameIdx = self.index(itNum, self.COL_NAME, parent)
            itName = self.itemFromIndex(itNameIdx)
            entry = itName.entry
            self.entryCache.register(entry)


class IPFSHashExplorerToolBox(GalacteekTab):
    """
    Organizes IPFSHashExplorerWidgets with a QToolBox
    """

    def __init__(self, gWindow, hashRef, maxItems=16, parent=None):
        super(IPFSHashExplorerToolBox, self).__init__(gWindow)

        self.rootHash = hashRef
        self.maxItems = maxItems

        self.toolbox = QToolBox()
        self.exLayout = QVBoxLayout()
        self.exLayout.addWidget(self.toolbox)

        self.vLayout.addLayout(self.exLayout)
        if self.rootHash:
            self.viewHash(self.rootHash)

    @property
    def itemsCount(self):
        return self.toolbox.count()

    def viewHash(self, hashRef, addClose=False, autoOpenFolders=False):
        w = self.lookup(hashRef)
        if w:
            self.toolbox.setCurrentWidget(w)
            return True

        if self.itemsCount > self.maxItems:
            return False

        view = IPFSHashExplorerWidget(hashRef,
                                      parent=self, addClose=addClose,
                                      autoOpenFolders=autoOpenFolders)
        view.closeRequest.connect(functools.partial(
            self.remove, view))
        view.directoryOpenRequest.connect(
            lambda cid: self.viewHash(cid, addClose=True))

        idx = self.toolbox.addItem(view, getIconIpfsWhite(), hashRef)
        self.toolbox.setCurrentIndex(idx)
        view.reFocus()
        return True

    def lookup(self, hashRef):
        for idx in range(self.itemsCount):
            if self.toolbox.itemText(idx) == hashRef:
                return self.toolbox.widget(idx)

    def remove(self, view):
        idx = self.toolbox.indexOf(view)
        if idx:
            self.toolbox.removeItem(idx)
            # Always display previous index in the stack
            if self.itemsCount > 0:
                rIdx = idx - 1
                view = self.toolbox.widget(rIdx)
                if view:
                    self.toolbox.setCurrentWidget(view)
                    view.reFocus()


class IPFSHashExplorerStack(GalacteekTab):
    """
    Organizes IPFSHashExplorerWidgets with a QStackedWidget
    """

    def __init__(self, gWindow, hashRef, maxItems=16, parent=None):
        super(IPFSHashExplorerStack, self).__init__(gWindow)

        self.rootHash = hashRef
        self.maxItems = maxItems

        self.stack = QStackedWidget(self)
        self.exLayout = QVBoxLayout()
        self.exLayout.addWidget(self.stack)

        self.vLayout.addLayout(self.exLayout)
        if self.rootHash:
            self.viewHash(self.rootHash)

    @property
    def itemsCount(self):
        return self.stack.count()

    def viewHash(self, hashRef, addClose=False, autoOpenFolders=False):
        view = IPFSHashExplorerWidget(hashRef,
                                      parent=self, addClose=addClose,
                                      showCidLabel=True,
                                      autoOpenFolders=autoOpenFolders)
        view.closeRequest.connect(functools.partial(
            self.remove, view))
        view.directoryOpenRequest.connect(
            lambda cid: self.viewHash(cid, addClose=True))

        self.stack.insertWidget(self.stack.count(), view)
        self.stack.setCurrentWidget(view)
        view.reFocus()
        return True

    def remove(self, view):
        try:
            self.stack.removeWidget(view)
        except:
            pass


class TreeEventFilter(QObject):
    copyPressed = pyqtSignal()
    returnPressed = pyqtSignal()
    backspacePressed = pyqtSignal()

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyPress:
            modifiers = event.modifiers()

            key = event.key()
            if key == Qt.Key_Return:
                self.returnPressed.emit()
                return True
            if key == Qt.Key_Backspace:
                self.backspacePressed.emit()
                return True
            if modifiers & Qt.ControlModifier:
                if key == Qt.Key_C:
                    self.copyPressed.emit()
                    return True
        return False


class MultihashTreeView(QTreeView):
    pass


class IPFSHashExplorerWidget(QWidget):
    closeRequest = pyqtSignal()
    directoryOpenRequest = pyqtSignal(str)
    fileOpenRequest = pyqtSignal(IPFSPath)
    parentMultihashSet = pyqtSignal(str)

    def __init__(self, hashRef, addClose=False,
                 mimeDetectionMethod='db',
                 addActions=True, autoOpenFiles=True,
                 showCidLabel=False,
                 autoOpenFolders=False, parent=None):
        super(IPFSHashExplorerWidget, self).__init__(parent)

        self.parent = parent
        self.app = QApplication.instance()
        self.gWindow = self.app.mainWindow
        self.mimeDetectionMethod = mimeDetectionMethod
        self.model = IPFSHashItemModel(self)

        self.parentMultihash = None
        self.parentButton = None
        self.rootHash = None
        self.changeMultihash(hashRef)

        self.mainLayout = QVBoxLayout(self)
        self.setLayout(self.mainLayout)

        self.autoOpenFolders = autoOpenFolders
        self.autoOpenFiles = autoOpenFiles
        self.hLayoutTop = QHBoxLayout()
        self.hLayoutInfo = QHBoxLayout()
        self.hLayoutCtrl = QHBoxLayout()
        self.hLayoutTop.addLayout(self.hLayoutInfo)
        spacerItem = QSpacerItem(10, 20, QSizePolicy.Expanding,
                                 QSizePolicy.Minimum)
        self.hLayoutTop.addItem(spacerItem)
        self.hLayoutTop.addLayout(self.hLayoutCtrl)

        self.labelInfo = QLabel()

        self.hLayoutInfo.addWidget(self.labelInfo, 0, Qt.AlignLeft)

        if addClose:
            self.closeButton = QPushButton('Close')
            self.closeButton.clicked.connect(self.onCloseView)
            self.closeButton.setMaximumWidth(100)
            self.closeButton.setShortcut(QKeySequence('Ctrl+w'))
            self.hLayoutCtrl.addWidget(self.closeButton, 0, Qt.AlignLeft)

        if addActions:
            self.addButtons()

        if showCidLabel:
            path = IPFSPath(self.rootHash, autoCidConv=True)
            labelMultihash = IPFSUrlLabel(path)
            clipButton = IPFSPathClipboardButton(path)
            layout = QHBoxLayout()
            layout.addWidget(labelMultihash)
            layout.addWidget(clipButton)
            layout.addItem(QSpacerItem(10, 20, QSizePolicy.Expanding,
                                       QSizePolicy.Minimum))
            self.mainLayout.addLayout(layout)

        self.mainLayout.addLayout(self.hLayoutTop)

        self.parentMultihashSet.connect(self.onParentMultihash)

        self.tree = MultihashTreeView()
        self.tree.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.tree.setDragDropMode(QAbstractItemView.DragOnly)
        self.mainLayout.addWidget(self.tree)

        self.initModel()

        self.tree.setContextMenuPolicy(Qt.CustomContextMenu)
        self.tree.customContextMenuRequested.connect(self.onContextMenu)
        self.tree.doubleClicked.connect(self.onDoubleClicked)
        self.tree.setSortingEnabled(True)

        self.evfilter = IPFSTreeKeyFilter(self.tree)
        self.evfilter.copyHashPressed.connect(self.onCopyItemHash)
        self.evfilter.copyPathPressed.connect(self.onCopyItemPath)
        self.evfilter.returnPressed.connect(self.onReturnPressed)
        self.evfilter.backspacePressed.connect(self.onBackspacePressed)
        self.tree.installEventFilter(self.evfilter)
        self.iconFolder = getIcon('folder-open.png')
        self.iconFile = getIcon('file.png')
        self.iconUnknown = getIcon('unknown-file.png')

        self.updateTree()

    def changeMultihash(self, cid):
        if cidValid(cid):
            self.parentMultihash = self.rootHash

            if self.parentMultihash is not None:
                self.parentMultihashSet.emit(self.parentMultihash)

            self.rootHash = cid
            self.rootPath = IPFSPath(self.rootHash, autoCidConv=True)
            self.cid = cidhelpers.getCID(self.rootHash)
            self.initModel()

    def goToParent(self):
        if self.parentMultihash:
            self.changeMultihash(self.parentMultihash)
            self.updateTree()

    def onParentMultihash(self, parent):
        pass

    def initModel(self):
        self.model.clearModel()
        self.model.setHorizontalHeaderLabels(
            [iFileName(), iFileSize(), iMimeType(), iCID()])
        self.itemRoot = self.model.invisibleRootItem()
        self.itemRootIdx = self.model.indexFromItem(self.itemRoot)

    def addButtons(self):
        self.getTask = None
        self.getButton = QPushButton(iDownload())
        self.getButton.clicked.connect(self.onGet)
        self.getLabel = QLabel()
        self.getProgress = QProgressBar()
        self.getProgress.setMinimum(0)
        self.getProgress.setMaximum(100)
        self.getProgress.hide()

        self.markButton = QPushButton(getIcon('hashmarks.png'), iHashmark())
        self.markButton.clicked.connect(self.onHashmark)

        self.pinButton = QPushButton(iPin())
        self.pinButton.clicked.connect(self.onPin)

        self.hLayoutCtrl.addWidget(self.getButton)
        self.hLayoutCtrl.addWidget(self.pinButton)
        self.hLayoutCtrl.addWidget(self.markButton)
        self.hLayoutCtrl.addWidget(self.getLabel)
        self.hLayoutCtrl.addWidget(self.getProgress)

    def setInfo(self, text):
        self.labelInfo.setText(text)

    def reFocus(self):
        self.tree.setFocus(Qt.OtherFocusReason)

    @ipfsOp
    async def gitClone(self, ipfsop, entry, dest):
        """
        Clone the git repository contained within entry, to directory dest
        """
        from git.repo import base
        from git.exc import InvalidGitRepositoryError

        self.gitButton.setEnabled(False)

        mHash = cidConvertBase32(entry['Hash'])
        getRet = await ipfsop.client.get(mHash,
                                         dstdir=self.app.tempDir.path())
        if not getRet:
            self.gitButton.setEnabled(True)
            return messageBox('Could not fetch the git repository')

        repoPath = os.path.join(self.app.tempDir.path(), mHash)
        try:
            repo = base.Repo(repoPath)
        except InvalidGitRepositoryError:
            return messageBox(iGitInvalid())

        dstPath = '{}.git'.format(
            os.path.join(dest, self.rootHash))

        # Clone it now. No need to run it in a threadpool since the git module
        # will run a git subprocess for the cloning
        try:
            repo.clone(dstPath)
        except Exception as e:
            self.gitButton.setEnabled(True)
            return messageBox(iGitErrorCloning(str(e)))

        messageBox(iGitClonedRepo(dstPath))
        self.gitButton.setEnabled(True)

    def addGitControl(self, gitEntry):
        """
        Adds a tool button for making operations on the git repo
        """
        def clone(entry):
            dirSel = directorySelect()
            if dirSel:
                self.app.task(self.gitClone, entry, dirSel)

        self.gitMenu = QMenu(parent=self)
        self.gitMenu.addAction(iGitClone(), lambda: clone(gitEntry))
        self.gitButton = QToolButton(self)
        self.gitButton.setText('Git')
        self.gitButton.setMenu(self.gitMenu)
        self.gitButton.setPopupMode(QToolButton.MenuButtonPopup)
        self.hLayoutCtrl.addWidget(self.gitButton)

    def onHashmark(self):
        addHashmark(self.app.marksLocal,
                    str(self.rootPath), '',
                    stats=self.app.ipfsCtx.objectStats.get(
                        self.rootPath.objPath, {}))

    def onGet(self):
        dirSel = directorySelect()
        if dirSel:
            self.getTask = self.app.task(self.getResource,
                                         self.rootPath.objPath,
                                         dirSel)

    def onPin(self):
        self.app.task(self.app.ipfsCtx.pinner.queue,
                      self.rootPath.objPath, True,
                      None)

    def onReturnPressed(self):
        currentIdx = self.tree.currentIndex()
        if currentIdx.isValid():
            self.onDoubleClicked(currentIdx)

    def onCopyItemHash(self):
        dataHash = self.model.getHashFromIdx(self.tree.currentIndex())
        self.app.setClipboardText(dataHash)

    def onCopyItemPath(self):
        nameItem = self.model.getNameItemFromIdx(self.tree.currentIndex())
        if nameItem:
            self.app.setClipboardText(nameItem.getFullPath())

    def onBackspacePressed(self):
        pass

    def onCloseView(self):
        self.closeRequest.emit()

    def browse(self, hash):
        self.gWindow.addBrowserTab().browseIpfsHash(hash)

    def browseFs(self, path):
        self.gWindow.addBrowserTab().browseFsPath(path)

    def onDoubleClicked(self, idx):
        nameItem = self.model.getNameItemFromIdx(idx)
        dataHash = self.model.getHashFromIdx(idx)

        if nameItem.isDir():
            self.directoryOpenRequest.emit(dataHash)
        elif nameItem.isFile() or nameItem.isUnknown():
            self.openFile(nameItem, dataHash)

    def openFile(self, item, fileHash):
        if item.mimeType is None or item.mimeType == 'text/html':
            return self.browse(fileHash)

        if self.autoOpenFiles:
            ensure(self.openWithRscOpener(item, fileHash))
        else:
            self.fileOpenRequest.emit(item.ipfsPath)

    @ipfsOp
    async def openWithRscOpener(self, ipfsop, item, fileHash):
        # Pass a null mimetype to the resource opener so that it
        # redetects the mimetype with a full (but slower) mime detection method
        opener = self.app.resourceOpener
        await opener.open(item.getFullPath(), None)

    def updateTree(self):
        if self.rootPath and self.rootPath.valid:
            self.app.task(self.listMultihash, self.rootPath.objPath,
                          parentItem=self.itemRoot)

    def onContextMenu(self, point):
        selModel = self.tree.selectionModel()
        rows = selModel.selectedRows()

        items = [self.model.getNameItemFromIdx(idx) for idx in rows]
        menu = QMenu(self)

        def pinRecursive():
            for item in items:
                fp = item.getFullPath()
                self.app.task(self.app.ipfsCtx.pinner.queue, fp, True, None)

        def download():
            dirSel = directorySelect()
            for item in items:
                self.app.task(self.getResource, item.getFullPath(),
                              dirSel)

        menu.addAction(getIcon('pin.png'), iPinRecursive(),
                       pinRecursive)
        menu.addAction(iDownload(), download)

        menu.exec(self.tree.mapToGlobal(point))

    async def timedList(self, ipfsop, objPath, parentItem, autoexpand,
                        secs, resolve_type):
        return await asyncio.wait_for(
            self.list(ipfsop, objPath,
                      parentItem=parentItem, autoexpand=autoexpand,
                      resolve_type=resolve_type), secs)

    @ipfsOp
    async def listMultihash(self, ipfsop, objPath, parentItem,
                            autoexpand=False):
        """ Lists contents of IPFS object referenced by objPath,
            and change the tree's model afterwards.

            We first try with resolve-type set to true, and if that gets a
            timeout do the same call but without resolving nodes types
        """

        try:
            self.setInfo(iLoading())
            await self.timedList(ipfsop, objPath, parentItem,
                                 autoexpand, 15, True)
        except asyncio.TimeoutError:
            self.setInfo(iTimeoutTryNoResolve())

            try:
                await self.timedList(ipfsop, objPath, parentItem,
                                     autoexpand, 10, False)
            except asyncio.TimeoutError:
                # That's a dead end .. bury that hash please ..
                self.setInfo(iTimeoutInvalidHash())
                return

        except aioipfs.APIError:
            self.setInfo(iErrNoCx())
            return

        self.tree.setModel(self.model)
        self.tree.header().setSectionResizeMode(self.model.COL_NAME,
                                                QHeaderView.ResizeToContents)
        self.tree.header().setSectionResizeMode(self.model.COL_MIME,
                                                QHeaderView.ResizeToContents)
        self.tree.sortByColumn(self.model.COL_NAME, Qt.AscendingOrder)

        if self.app.settingsMgr.hideHashes:
            self.tree.hideColumn(self.model.COL_HASH)

        rStat = await ipfsop.objStat(objPath)
        statInfo = StatInfo(rStat)

        if statInfo.valid and self.cid:
            self.setInfo(iCIDInfo(self.cid.version,
                                  statInfo.numLinks,
                                  sizeFormat(statInfo.totalSize)))

    async def list(self, op, path, parentItem=None,
                   autoexpand=False, resolve_type=True):

        parentItemSibling = self.model.sibling(parentItem.row(),
                                               self.model.COL_HASH,
                                               parentItem.index())
        parentItemHash = self.model.data(parentItemSibling)
        if parentItemHash is None:
            parentItemHash = self.rootHash

        async for obj in op.list(path, resolve_type):
            for entry in obj['Links']:
                await op.sleep()

                cid = cidConvertBase32(entry['Hash'])
                if cid in self.model.entryCache:
                    continue

                if len(entry['Name']) > 32:
                    entryName = entry['Name'][0:32]
                else:
                    entryName = entry['Name']

                nItemName = IPFSNameItem(entry, entryName, None)

                if self.mimeDetectionMethod == 'db':
                    nItemName.mimeFromDb(self.app.mimeDb)
                elif self.mimeDetectionMethod == 'magic':
                    mType = await detectMimeType(cid)
                    if mType:
                        nItemName.mimeType = str(mType)

                nItemName.setParentHash(parentItemHash)
                nItemSize = IPFSItem(sizeFormat(entry['Size']))
                nItemSize.setToolTip(str(entry['Size']))
                nItemMime = IPFSItem(nItemName.mimeType or iUnknown())
                nItemHash = IPFSItem(cid)
                nItemHash.setToolTip(cidInfosMarkup(cid))
                nItemName.setToolTip(entry['Name'])

                if nItemName.isDir():
                    nItemName.setIcon(self.iconFolder)
                elif nItemName.isFile():
                    if nItemName.mimeType:
                        mIcon = getMimeIcon(nItemName.mimeType)
                        if mIcon:
                            nItemName.setIcon(mIcon)
                        else:
                            nItemName.setIcon(self.iconFile)
                    else:
                        nItemName.setIcon(self.iconFile)
                elif nItemName.isUnknown():
                    nItemName.setIcon(self.iconUnknown)

                nItem = [nItemName, nItemSize, nItemMime, nItemHash]
                parentItem.appendRow(nItem)

                if nItemName.isDir() and self.autoOpenFolders:
                    # Automatically open sub folders. Used by unit tests
                    self.directoryOpenRequest.emit(dataHash)

                if nItemName.isDir() and entry['Name'] == '.git':
                    # If there's a git repo here, add a control button
                    self.addGitControl(entry)

    @ipfsStatOp
    async def getResource(self, ipfsop, rPath, dest, rStat):
        """
        Get the resource referenced by rPath to directory dest
        """

        statInfo = StatInfo(rStat)
        if not statInfo.valid:
            return

        self.getProgress.show()

        async def onGetProgress(ref, bytesRead, arg):
            per = int((bytesRead * 100) / statInfo.totalSize)
            self.getLabel.setText('Downloaded: {0}'.format(
                sizeFormat(bytesRead)))
            self.getProgress.setValue(per)

        await ipfsop.client.get(rPath, dstdir=dest,
                                progress_callback=onGetProgress,
                                chunk_size=32768)

        self.getLabel.setText('Download finished')
        self.getProgress.hide()


class IPFSHashExplorerWidgetFollow(IPFSHashExplorerWidget):
    def __init__(self, hashRef, parent=None, **kwargs):
        super(IPFSHashExplorerWidgetFollow, self).__init__(
            hashRef, parent=parent, **kwargs)

        self.directoryOpenRequest.connect(self.onOpenDir)

    def onOpenDir(self, cid):
        self.changeMultihash(cid)
        self.updateTree()

    def onBackspacePressed(self):
        self.goToParent()



"""
Protobuf definitions for IPFS
"""

import base64


UNIXFS_DT_RAW = 0
UNIXFS_DT_DIRECTORY = 1
UNIXFS_DT_FILE = 2
UNIXFS_DT_METADATA = 3
UNIXFS_DT_SYMLINK = 4
UNIXFS_DT_HAMTSHARD = 5


unixfsDtNames = {
    UNIXFS_DT_DIRECTORY: 'directory',
    UNIXFS_DT_FILE: 'file',
    UNIXFS_DT_HAMTSHARD: 'hamtshard',
    UNIXFS_DT_METADATA: 'metadata',
    UNIXFS_DT_RAW: 'raw',
    UNIXFS_DT_SYMLINK: 'symlink'
}


def decodeUnixfsDagNode(data):
    """
    Import unixfs_pb2 from the function for now, it's rather heavy
    to import protobuf on startup
    """

    from . import unixfs_pb2
    from .protobuf_to_dict import protobuf_to_dict

    try:
        message = unixfs_pb2.Data()
        message.MergeFromString(
            base64.b64decode(data))
        obj = protobuf_to_dict(message)
    except Exception:
        return None

    return {
        'type': obj['Type'],
        'size': obj.get('filesize'),
        'blocksizes': obj.get('blocksizes'),
        'data': obj.get('Data')
    }



from html.parser import HTMLParser
from urllib.parse import unquote

from PyQt5.QtCore import QUrl

from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.core.schemes import isIpfsUrl


class IPFSLinksParser(HTMLParser):
    """
    Primitive HTML IPFS links parser, used to scan IPFS URLs in
    HTML pages rendered by the browser
    """
    def __init__(self, basePath):
        """
        :param IPFSPath basePath: The page's base IPFS path
        """

        super().__init__()
        self.links = []
        self.base = basePath

    def handle_starttag(self, tag, alist):
        if tag not in ['a', 'link']:
            return

        attrs = dict(alist)
        href = attrs.get('href')

        if not isinstance(href, str):
            return

        if len(href) > 1024:
            return

        href = unquote(href)

        url = QUrl(href)
        if not url.isValid():
            return

        if not url.scheme() or url.isRelative():
            # Relative URL
            p = self.base.child(href)
            if p and self.pathValid(p):
                self.links.append(p)

        elif isIpfsUrl(url):
            # Absolute URL
            p = IPFSPath(url.toString())
            if self.pathValid(p):
                self.links.append(p)

    def pathValid(self, p):
        return p.valid and p != self.base

    def handle_endtag(self, tag):
        pass

    def handle_data(self, data):
        pass



import json
import asyncio

from PyQt5.QtWidgets import QTextEdit
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QCheckBox
from PyQt5.QtWidgets import QPushButton

from PyQt5.QtCore import Qt
from PyQt5.QtCore import QSaveFile
from PyQt5.QtCore import QIODevice

from galacteek import log
from galacteek.ipfs.wrappers import ipfsOp
from galacteek.ui.widgets import GalacteekTab
from galacteek.ui.helpers import saveFileSelect, messageBox
from galacteek.core import jtraverse


class EventLogWidget(GalacteekTab):
    """
    Widget to display IPFS log events
    """

    def __init__(self, gWindow):
        super().__init__(gWindow)

        self.logZone = QTextEdit()
        self.logZone.setReadOnly(True)

        self.saveButton = QPushButton('Save')
        self.saveButton.clicked.connect(self.onSave)
        self.clearButton = QPushButton('Clear')
        self.clearButton.clicked.connect(lambda: self.logZone.clear())

        layout = QVBoxLayout()
        hLayout = QHBoxLayout()

        self.checkCore = QCheckBox('Core events')
        self.checkCore.setCheckState(Qt.Checked)
        self.checkPubsub = QCheckBox('Pubsub events')
        self.checkDht = QCheckBox('DHT events')
        self.checkBitswap = QCheckBox('Bitswap events')
        self.checkAll = QCheckBox('All events')
        self.checkAll.stateChanged.connect(self.onCheckAll)

        hLayout.addWidget(self.checkCore)
        hLayout.addWidget(self.checkDht)
        hLayout.addWidget(self.checkPubsub)
        hLayout.addWidget(self.checkBitswap)
        hLayout.addWidget(self.checkAll)
        hLayout.addWidget(self.clearButton)
        hLayout.addWidget(self.saveButton)

        layout.addLayout(hLayout)
        layout.addWidget(self.logZone)

        self.tskLog = self.app.task(self.logWatch)
        self.vLayout.addLayout(layout)

    def onCheckAll(self, state):
        self.checkCore.setEnabled(not state)
        self.checkDht.setEnabled(not state)
        self.checkBitswap.setEnabled(not state)
        self.checkPubsub.setEnabled(not state)

    def onSave(self):
        fPath = saveFileSelect()
        if fPath:
            file = QSaveFile(fPath)

            if not file.open(QIODevice.WriteOnly):
                return messageBox('Cannot open file for writing')

            file.write(self.logZone.toPlainText().encode())
            file.commit()

    def displayEvent(self, event):
        self.logZone.append(json.dumps(event, indent=4))

    @ipfsOp
    async def logWatch(self, op):
        try:
            await op.client.log.level('all', 'info')

            async for event in op.client.log.tail():
                display = False
                parser = jtraverse.traverseParser(event)

                systems = [parser.traverse('system'),
                           parser.traverse('Tags.system')]

                if ('core' in systems or 'addrutil' in systems) and \
                        self.checkCore.isChecked():
                    display = True
                elif 'dht' in systems and self.checkDht.isChecked():
                    display = True
                elif 'bitswap' in systems and self.checkBitswap.isChecked():
                    display = True
                elif 'pubsub' in systems and self.checkPubsub.isChecked():
                    display = True

                if self.checkAll.isChecked():
                    display = True

                if display is True:
                    self.displayEvent(event)

                await op.sleep()

        except asyncio.CancelledError:
            return
        except Exception:
            log.debug('Unknown error ocurred while reading ipfs log')

    def onClose(self):
        self.tskLog.cancel()
        return True



from galacteek.ipfs.wrappers import ipfsOp  # noqa
from galacteek.ipfs.wrappers import ipfsOpFn  # noqa


def kilobytes(kb):
    return (1024 * kb)


def megabytes(mb):
    return (1024 * kilobytes(mb))



import asyncio
import async_timeout
import uuid

from PyQt5.QtWebEngineWidgets import QWebEnginePage
from PyQt5.QtWebEngineWidgets import QWebEngineView
from PyQt5.QtWebEngineWidgets import QWebEngineProfile
from PyQt5.QtWebChannel import QWebChannel

from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QUrl
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QRegularExpression

from PyQt5.QtCore import QVariant

from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QVBoxLayout

from PyQt5.QtGui import QSyntaxHighlighter
from PyQt5.QtGui import QTextCharFormat
from PyQt5.QtGui import QFont

from galacteek.ipfs.cidhelpers import joinIpfs
from galacteek.ipfs.cidhelpers import stripIpfs
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.ipfsops import *
from galacteek.ipfs.wrappers import ipfsOp
from galacteek.ipfs import ipfssearch
from galacteek.dweb.render import renderTemplate
from galacteek.core.analyzer import ResourceAnalyzer
from galacteek import ensure
from galacteek import log

from .helpers import *
from .hashmarks import *
from .widgets import *
from .i18n import *


def iResultsInfo(count, maxScore):
    return QCoreApplication.translate(
        'IPFSSearchResults',
        'Results count: <b>{0}</b> (max score: <b>{1}</b>)').format(
        count,
        maxScore)


def iNoResults():
    return QCoreApplication.translate('IPFSSearchResults',
                                      '<b>No results found</b>')


def iErrFetching():
    return QCoreApplication.translate('IPFSSearchResults',
                                      '<b>Error while fetching results</b>')


def iSearching():
    return QCoreApplication.translate('IPFSSearchResults',
                                      '<b>Searching ...</b>')


class IPFSSearchButton(QToolButton):
    hovered = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.iconNormal = getIcon('search-engine.png')
        self.iconActive = getIcon('search-engine-zoom.png')

    def enterEvent(self, ev):
        self.hovered.emit()


class Highlighter(QSyntaxHighlighter):
    """
    Highlights the search query string in the results
    """

    def __init__(self, highStr, doc):
        super().__init__(doc)

        self.highStr = highStr
        self.fmt = QTextCharFormat()
        self.fmt.setFontWeight(QFont.Bold)
        self.fmt.setForeground(Qt.red)

    def highlightBlock(self, text):
        rExpr = QRegularExpression(self.highStr)
        matches = rExpr.globalMatch(text)

        if not matches.isValid():
            return

        while matches.hasNext():
            match = matches.next()
            if not match.isValid():
                break
            self.setFormat(match.capturedStart(), match.capturedLength(),
                           self.fmt)


class NoResultsPage(QWebEnginePage):
    def __init__(self, tmpl, parent=None):
        super(NoResultsPage, self).__init__(parent)
        self.setHtml('<html><body><p>NO RESULTS</p></body></html>')


class BaseSearchPage(QWebEnginePage):
    pageRendered = pyqtSignal(str)

    def __init__(self, parent, profile=None):
        if profile:
            super(BaseSearchPage, self).__init__(profile, parent)
        else:
            super(BaseSearchPage, self).__init__(parent)

        self.pageRendered.connect(self.onPageRendered)

    def onPageRendered(self, html):
        self.setHtml(html, baseUrl=QUrl('qrc:/'))


class SearchInProgressPage(BaseSearchPage):
    def __init__(self, tmpl, parent=None):
        super(SearchInProgressPage, self).__init__(parent)
        self.template = tmpl
        ensure(self.render())

    async def render(self):
        html = await renderTemplate(self.template)
        if html:
            self.pageRendered.emit(html)


class SearchResultsPageFactory(QObject):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.app = QApplication.instance()
        self.resultsTemplate = self.app.getJinjaTemplate('ipfssearch.html')

        self.webProfile = QWebEngineProfile()
        self.installScripts()

        self._pages = {}
        self.genPages(3)

    def installScripts(self):
        self.webScripts = self.webProfile.scripts()

        exSc = self.webScripts.findScript('ipfs-http-client')
        if self.app.settingsMgr.jsIpfsApi is True and exSc.isNull():
            for script in self.app.scriptsIpfs:
                self.webScripts.insert(script)

    def genPages(self, count):
        [self.makePage() for i in range(0, count)]

    def getPage(self):
        for uid, data in self._pages.items():
            page = data['page']
            if not page.used:
                page.used = True
                self.app.loop.call_soon(self.genPages, 2)
                return page, data['handler'], data['channel']

    def makePage(self):
        uid = str(uuid.uuid4())
        handler = IPFSSearchHandler(self)
        channel = QWebChannel()

        self._pages[uid] = {
            'channel': channel,
            'page': None,
            'handler': handler
        }

        channel.registerObject('ipfssearch', handler)

        resultsPage = SearchResultsPage(
            self.webProfile,
            handler,
            self.resultsTemplate,
            None,
            self.app.getIpfsConnectionParams(),
            parent=self,
            webchannel=self._pages[uid]['channel']
        )

        self._pages[uid]['page'] = resultsPage
        return resultsPage


class SearchResultsPage(BaseSearchPage):
    def __init__(
            self,
            profile,
            handler,
            tmplMain,
            tmplHits,
            ipfsConnParams,
            webchannel=None,
            parent=None):
        super(SearchResultsPage, self).__init__(parent, profile=profile)

        self.used = False

        if webchannel:
            self.setWebChannel(webchannel)

        self.app = QApplication.instance()

        self.url = QUrl('qrc:/')
        self.handler = handler
        self.template = tmplMain
        self.templateHits = tmplHits
        self.ipfsConnParams = ipfsConnParams
        self.noStatTimeout = 12
        ensure(self.render())

    def javaScriptConsoleMessage(self, level, message, lineNumber, sourceId):
        log.info(
            'JS: level: {0}, source: {1}, line: {2}, message: {3}'.format(
                level,
                sourceId,
                lineNumber,
                message))

    async def render(self):
        html = await renderTemplate(self.template,
                                    ipfsConnParams=self.ipfsConnParams)
        self.setHtml(html, baseUrl=QUrl('qrc:/'))


class IPFSSearchHandler(QObject):
    ready = pyqtSignal()

    resultsReceived = pyqtSignal(ipfssearch.IPFSSearchResults, bool)

    resultReady = pyqtSignal(str, QVariant)

    objectReady = pyqtSignal(str)
    objectStatAvailable = pyqtSignal(str, dict, dict)
    objectStatUnavailable = pyqtSignal(str)

    filtersChanged = pyqtSignal()
    clear = pyqtSignal()
    resetForm = pyqtSignal()
    vPageChanged = pyqtSignal(int)
    vPageStatus = pyqtSignal(int, int)

    searchTimeout = pyqtSignal(int)
    searchError = pyqtSignal()
    searchCancelled = pyqtSignal()
    searchStarted = pyqtSignal(str)
    searchComplete = pyqtSignal()

    def __init__(self, parent):
        super().__init__(parent)

        self.app = QApplication.instance()
        self.analyzer = ResourceAnalyzer(self)
        self.noStatTimeout = 8

        self.vPageCurrent = 0
        self.pagesPerVpage = 2
        self.pageCount = 0
        self.searchQuery = ''

        self.filters = {}
        self._tasks = []
        self._taskSearch = None
        self._cResults = []

    @property
    def vPageCurrent(self):
        return self._vPageCurrent

    @vPageCurrent.setter
    def vPageCurrent(self, page):
        self._vPageCurrent = page
        self.vPageChanged.emit(self.vPageCurrent)

    def setSearchWidget(self, widget):
        self.searchW = widget

    def reload(self):
        self.vPageCurrent = 0
        self.filtersChanged.emit()

    def init(self):
        self.vPageCurrent = 0

    def formReset(self):
        self.resetForm.emit()

    def cleanup(self):
        self._cancelTasks()

        self.vPageCurrent = 0
        self.clear.emit()
        self._cResults = []

    def _cancelTasks(self):
        try:
            if self._taskSearch:
                self._taskSearch.cancel()

            for task in self._tasks:
                task.cancel()
        except Exception:
            log.debug('Failed to cancel search tasks')

        self._tasks = []

    def spawnSearchTask(self):
        self._taskSearch = ensure(self.runSearch(self.searchQuery))

    @pyqtSlot()
    def cancelSearchTasks(self):
        self._cancelTasks()

    @pyqtSlot(str)
    def fetchObject(self, path):
        ensure(self.fetch(path))

    @pyqtSlot(str)
    def pinObject(self, path):
        ensure(self.pin(path))

    @ipfsOp
    async def pin(self, op, path):
        pinner = op.ctx.pinner
        await pinner.queue(path, True, None, qname='ipfs-search')

    @pyqtSlot()
    def previousPage(self):
        if self.vPageCurrent > 0:
            self._cancelTasks()
            self.clear.emit()
            self.vPageCurrent -= 1
            self.spawnSearchTask()

    @pyqtSlot()
    def nextPage(self):
        self._cancelTasks()
        self.clear.emit()
        self.vPageCurrent += 1
        self.spawnSearchTask()

    @ipfsOp
    async def fetch(self, op, path):
        import tempfile
        try:
            data = await asyncio.wait_for(
                op.client.cat(path), 5)
        except BaseException:
            pass
        else:
            file = tempfile.NamedTemporaryFile(delete=False)
            file.write(data)
            self.objectReady.emit(file.name)

    @pyqtSlot(str)
    def openLink(self, path):
        ensure(self.app.resourceOpener.open(path))

    @pyqtSlot(str)
    def clipboardInput(self, path):
        if isinstance(path, str):
            self.app.setClipboardText(path)

    def findByHash(self, mHash):
        for results in self._cResults:
            found = results.findByHash(mHash)
            if found:
                return found

    @pyqtSlot(str)
    def hashmark(self, path):
        hashV = stripIpfs(path)
        hit = self.findByHash(hashV)

        if not hit:
            return

        title = hit.get('title', iUnknown()) if hit else ''
        descr = hit.get('description', iUnknown()) if hit else ''
        type = hit.get('type', None)
        pinSingle = (type == 'file')
        pinRecursive = (type == 'directory')

        addHashmark(self.app.marksLocal,
                    path, title, description=descr,
                    pin=pinSingle, pinRecursive=pinRecursive)

    @pyqtSlot(str)
    def explore(self, path):
        hashV = stripIpfs(path)
        if hashV:
            self.app.mainWindow.explore(hashV)

    @pyqtSlot(str, str)
    def search(self, searchQuery, cType):
        self.cleanup()
        self.searchQuery = searchQuery
        self.searchStarted.emit(self.searchQuery)
        self.filters = self.getFilters(cType)

        self.spawnSearchTask()

    def getFilters(self, cTypeS):
        filters = {}

        cType = cTypeS.lower()

        if cType == 'images':
            filters['metadata.Content-Type'] = 'image*'
        elif cType == 'videos':
            filters['metadata.Content-Type'] = 'video*'
        elif cType == 'music':
            filters['metadata.Content-Type'] = 'audio*'
        elif cType == 'text':
            filters['metadata.Content-Type'] = 'text*'

        return filters

    def sendHit(self, hit):
        hitHash = hit.get('hash', None)
        mimeType = hit.get('mimetype', None)

        ipfsPath = IPFSPath(hitHash, autoCidConv=True)
        if not ipfsPath.valid:
            return

        sizeFormatted = sizeFormat(hit.get('size', 0))

        pHit = {
            'hash': hitHash,
            'path': str(ipfsPath),
            'url': ipfsPath.ipfsUrl,
            'mimetype': mimeType if mimeType else '',
            'title': hit.get('title', iUnknown()),
            'size': hit.get('size', iUnknown()),
            'sizeformatted': sizeFormatted,
            'description': hit.get('description', None),
            'type': hit.get('type', iUnknown()),
            'first-seen': hit.get('first-seen', iUnknown())
        }

        self.resultReady.emit(hitHash, QVariant(pHit))

    async def fetchObjectStat(self, ipfsop, hit):
        cid = hit['hash']
        path = joinIpfs(cid)

        stat = await ipfsop.objStatCtxUpdate(path, timeout=self.noStatTimeout)
        if stat:
            self.objectStatAvailable.emit(cid, stat, hit)

    @ipfsOp
    async def runSearch(self, ipfsop, searchQuery, timeout=60):
        pageStart = self.vPageCurrent * self.pagesPerVpage
        pageCount = 0
        gotResults = False
        statusEmitted = False

        try:
            with async_timeout.timeout(timeout):
                async for sr in ipfssearch.search(
                        searchQuery,
                        pageStart=pageStart,
                        preloadPages=self.pagesPerVpage,
                        filters=self.filters,
                        sslverify=self.app.sslverify):
                    gotResults = True

                    await asyncio.sleep(0)
                    pageCount = sr.pageCount

                    if not isinstance(pageCount, int) or pageCount <= 0:
                        break

                    vpCount = pageCount / self.pagesPerVpage

                    if not statusEmitted:
                        self.vPageStatus.emit(self.vPageCurrent + 1,
                                              vpCount if vpCount > 0 else 1)
                        statusEmitted = True

                    self._cResults.append(sr)
                    self.pageCount = pageCount

                    for hit in sr.hits:
                        self.sendHit(hit)

                        await asyncio.sleep(0.2)

                        self._tasks.append(
                            ensure(self.fetchObjectStat(ipfsop, hit)))
        except asyncio.TimeoutError:
            self.searchTimeout.emit(timeout)
            return
        except asyncio.CancelledError:
            self.searchCancelled.emit()
            return
        except Exception as e:
            log.debug(
                'IPFSSearch: unknown exception while searching: {}'.format(
                    str(e)))

        if not gotResults:
            self.searchError.emit()
        else:
            self.searchComplete.emit()


class IPFSSearchView(QWidget):
    resultsReceived = pyqtSignal(ipfssearch.IPFSSearchResults, bool)
    titleNeedUpdate = pyqtSignal(str)

    def __init__(self, searchQuery='', parent=None):
        super(IPFSSearchView, self).__init__(parent)
        self.setLayout(QVBoxLayout())

        self.tab = parent

        self.app = QApplication.instance()
        self.setAttribute(Qt.WA_DeleteOnClose)

        # Templates
        self.resultsTemplate = self.app.getJinjaTemplate('ipfssearch.html')
        self.hitsTemplate = self.app.getJinjaTemplate('ipfssearch-hits.html')
        self.loadingTemplate = self.app.getJinjaTemplate(
            'ipfssearch-loading.html')

        self.browser = QWebEngineView(parent=self)
        self.layout().addWidget(self.browser)

        self.resultsPage, self.handler, self.channel = \
            self.app.mainWindow.ipfsSearchPageFactory.getPage()
        self.resultsPage.setParent(self)

        self.handler.searchStarted.connect(
            lambda query: self.titleNeedUpdate.emit(query))

        self.browser.setPage(self.resultsPage)
        self.browser.setFocus(Qt.OtherFocusReason)

        self.handler.init()


class IPFSSearchTab(GalacteekTab):
    def __init__(self, gWindow, query=None):
        super(IPFSSearchTab, self).__init__(gWindow)

        self.view = IPFSSearchView(query, parent=self)
        self.view.titleNeedUpdate.connect(
            lambda text: self.setTabName(iIpfsSearchText(text[0:12])))
        self.addToLayout(self.view)

    def onClose(self):
        self.view.handler.cleanup()
        self.view.handler.formReset()
        self.view.resultsPage.used = False
        del self.view
        return True



import os
import os.path
import asyncio
import re
import json
import signal

from galacteek import log


async def shell(arg):
    p = await asyncio.create_subprocess_shell((arg),
                                              stdin=asyncio.subprocess.PIPE,
                                              stdout=asyncio.subprocess.PIPE)

    stdout, stderr = await p.communicate()
    return stdout.decode()


async def ipfsConfig(binPath, param, value):
    return await shell("'{0}' config '{1}' '{2}'".format(
        binPath, param, value))


async def ipfsConfigProfileApply(binPath, profile):
    return await shell("'{0}' config profile apply {1}".format(
        binPath, profile))


async def ipfsConfigJson(binPath, param, value):
    return await shell("'{0}' config --json {1} {2}".format(
        binPath, param, json.dumps(value)))


async def ipfsConfigGetJson(binPath, param):
    return await shell("'{0}' config --json '{1}'".format(binPath, param))


class IPFSDProtocol(asyncio.SubprocessProtocol):
    """
    IPFS daemon process protocol

    This handles output from the IPFS daemon
    """

    def __init__(self, loop, exitFuture, startedFuture):
        self._loop = loop
        self._output = bytearray()
        self.eventStarted = asyncio.Event()
        self.exitFuture = exitFuture
        self.startedFuture = startedFuture
        self.apiStarted = False
        self.gatewayStarted = False
        self.swarmStarted = False
        self.daemonReady = False
        self.errAlreadyRunning = False

    @property
    def loop(self):
        return self._loop

    @property
    def output(self):
        return self._output

    def pipe_data_received(self, fd, data):
        try:
            msg = data.decode().strip()
        except BaseException:
            return

        # The output we expect might be different in earlier versions
        # i've mainly tested with go-ipfs > 0.4.11
        # TODO: implement ipfs process supervisor independent of daemon output

        for line in msg.split('\n'):
            if re.search('Error: ipfs daemon is running', line):
                self.errAlreadyRunning = True
            if re.search('Gateway.*server listening on', line):
                self.gatewayStarted = True
            if re.search('Swarm listening on', line):
                self.swarmStarted = True
            if re.search('API server listening', line):
                self.apiStarted = True
            if re.search('Daemon is ready', line):
                self.daemonReady = True
            if re.search('Error:', line):
                pass

        if self.daemonReady is True:
            if not self.startedFuture.done():
                self.startedFuture.set_result(True)
            self.eventStarted.set()

        self._output.extend(data)

    def process_exited(self):
        self.exitFuture.set_result(True)


DEFAULT_APIPORT = 5001
DEFAULT_SWARMPORT = 4001
DEFAULT_GWPORT = 8080


class AsyncIPFSDaemon(object):
    """
    Async IPFS daemon subprocess wrapper

    :param str repopath: IPFS repository path
    :param int apiport: API port number to listen on
    :param int swarmport: Swarm port number to listen on
    :param int gatewayport: HTTP GW port number to listen on
    :param int swarmLowWater: min swarm connections
    :param int swarmHighWater: max swarm connections
    :param int storageMax: max repository storage capacity, in GB
    :param bool pubsubEnable: enable pubsub
    :param bool noBootstrap: empty bootstrap
    :param bool corsEnable: enable CORS
    :param bool p2pStreams: enable P2P streams support
    :param bool gwWritable: make the HTTP gateway writable
    """

    def __init__(self, repopath, goIpfsPath='ipfs',
                 apiport=DEFAULT_APIPORT,
                 swarmport=DEFAULT_SWARMPORT,
                 gatewayport=DEFAULT_GWPORT, initRepo=True,
                 swarmLowWater=10, swarmHighWater=20, nice=20,
                 pubsubEnable=False, noBootstrap=False, corsEnable=True,
                 pubsubRouter='floodsub', namesysPubsub=False,
                 p2pStreams=False, migrateRepo=False, routingMode='dht',
                 gwWritable=False, storageMax=20, debug=False, loop=None):

        self.loop = loop if loop else asyncio.get_event_loop()
        self.exitFuture = asyncio.Future(loop=self.loop)
        self.startedFuture = asyncio.Future(loop=self.loop)
        self.repopath = repopath
        self.goIpfsPath = goIpfsPath
        self.apiport = apiport
        self.gatewayport = gatewayport
        self.swarmport = swarmport
        self.swarmLowWater = swarmLowWater
        self.swarmHighWater = swarmHighWater
        self.storageMax = storageMax
        self.initRepo = initRepo
        self.namesysPubsub = namesysPubsub
        self.pubsubEnable = pubsubEnable
        self.pubsubRouter = pubsubRouter
        self.corsEnable = corsEnable
        self.p2pStreams = p2pStreams
        self.noBootstrap = noBootstrap
        self.migrateRepo = migrateRepo
        self.gwWritable = gwWritable
        self.routingMode = routingMode
        self.nice = nice
        self.debug = debug

    async def start(self):
        # Set the IPFS_PATH environment variable
        os.environ['IPFS_PATH'] = self.repopath

        if not os.path.isdir(self.repopath):
            os.mkdir(self.repopath)

        if not os.path.exists(os.path.join(self.repopath, 'config')) or \
                not os.path.isdir(os.path.join(self.repopath, 'datastore')):
            # Pretty sure this is an empty repository path
            await shell('ipfs init')

        apifile = os.path.join(self.repopath, 'api')
        if os.path.exists(apifile):
            os.unlink(apifile)

        # Change the addresses/ports we listen on
        await ipfsConfig(self.goIpfsPath, 'Addresses.API',
                         '/ip4/127.0.0.1/tcp/{0}'.format(self.apiport))
        await ipfsConfig(self.goIpfsPath, 'Addresses.Gateway',
                         '/ip4/127.0.0.1/tcp/{0}'.format(self.gatewayport))

        # Swarm multiaddrs (ipv4 and ipv6)
        swarmAddrs = [
            "/ip4/0.0.0.0/tcp/{swarmport}".format(swarmport=self.swarmport),
            "/ip6/::/tcp/{swarmport}".format(swarmport=self.swarmport)
        ]

        await ipfsConfigJson(self.goIpfsPath, 'Addresses.Swarm',
                             json.dumps(swarmAddrs))

        # Swarm connection manager parameters
        await ipfsConfigJson(self.goIpfsPath, 'Swarm.ConnMgr.LowWater',
                             self.swarmLowWater)
        await ipfsConfigJson(self.goIpfsPath, 'Swarm.ConnMgr.HighWater',
                             self.swarmHighWater)
        await ipfsConfig(self.goIpfsPath, 'Swarm.ConnMgr.GracePeriod',
                         '60s')

        await ipfsConfig(self.goIpfsPath, 'Routing.Type', self.routingMode)

        if self.pubsubRouter in ['floodsub', 'gossipsub']:
            await ipfsConfig(self.goIpfsPath, 'Pubsub.Router',
                             self.pubsubRouter)

        await ipfsConfigJson(self.goIpfsPath,
                             'Swarm.DisableBandwidthMetrics', 'true')

        # Maximum storage
        await ipfsConfig(self.goIpfsPath, 'Datastore.StorageMax',
                         '{0}GB'.format(self.storageMax))

        # P2P streams
        if self.p2pStreams:
            await ipfsConfigJson(self.goIpfsPath,
                                 'Experimental.Libp2pStreamMounting',
                                 'true'
                                 )

        # CORS
        if self.corsEnable:
            # Setup the CORS headers, only allowing the gateway's origin
            await ipfsConfigJson(
                self.goIpfsPath,
                'API.HTTPHeaders.Access-Control-Allow-Credentials',
                '["true"]'
            )
            await ipfsConfigJson(
                self.goIpfsPath,
                'API.HTTPHeaders.Access-Control-Allow-Methods',
                '["GET", "POST"]'
            )
            await ipfsConfigJson(
                self.goIpfsPath,
                'API.HTTPHeaders.Access-Control-Allow-Origin',
                '["*"]'
            )

        if self.noBootstrap:
            await ipfsConfigJson(self.goIpfsPath, 'Bootstrap', '[]')

        await ipfsConfigJson(self.goIpfsPath, 'Gateway.Writable',
                             self.gwWritable)

        args = [self.goIpfsPath, 'daemon']

        if self.pubsubEnable:
            args.append('--enable-pubsub-experiment')

        if self.namesysPubsub:
            args.append('--enable-namesys-pubsub')

        if self.migrateRepo:
            args.append('--migrate')

        f = self.loop.subprocess_exec(
            lambda: IPFSDProtocol(self.loop, self.exitFuture,
                                  self.startedFuture),
            *args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)
        self.transport, self.proto = await f

        proc_pid = self.transport.get_pid()
        self.setProcLimits(proc_pid, nice=self.nice)
        return True

    def setProcLimits(self, pid, nice=20):
        try:
            import psutil
        except ImportError:
            return

        log.debug('Applying limits to process: {pid}'.format(pid=pid))

        try:
            proc = psutil.Process(pid)
            proc.nice(nice)
        except Exception:
            log.debug('Could not apply limits to process {pid}'.format(
                pid=pid))

    def stop(self):
        log.debug('Stopping IPFS daemon')
        try:
            self.transport.send_signal(signal.SIGINT)
            self.transport.send_signal(signal.SIGHUP)
            return True
        except Exception as e:
            self.terminateException = e
            return False



from yarl import URL
import aiohttp
import asyncio
import async_timeout
import re

from galacteek import log
from galacteek.ipfs.cidhelpers import IPFSPath

"""
This module was used by the ENSWhoisSchemeHandler for resolving
the content hash of ENS domains using the whoisens.org API
EthDNS will be the default resolver probably now
"""


whoisApiHost = 'api.whoisens.org'
whoisBaseUrl = URL.build(host=whoisApiHost,
                         port=443,
                         scheme='https')


async def ensContentHash(domain, sslverify=True, timeout=8):
    """
    Resolve the content hash of an ENS domain with enswhois.org

    :rtype: IPFSPath
    """

    url = whoisBaseUrl.join(URL('resolve/contenthash/{domain}'.format(
        domain=domain)))

    try:
        with async_timeout.timeout(timeout):
            async with aiohttp.ClientSession() as session:
                async with session.get(str(url), verify_ssl=sslverify) as resp:
                    data = await resp.json()
                    addr = data['result']['result']

                    if isinstance(addr, str) and addr.startswith('ipfs://'):
                        match = re.search(
                            'ipfs://(?P<cid>[A-Za-z0-9]+)$', addr)
                        if match:
                            return IPFSPath(match.group('cid'),
                                            autoCidConv=True)
    except asyncio.TimeoutError:
        return None
    except Exception:
        log.debug('ensContentHash: Error occured while resolving')
        return None


async def ensResolveAddr(domain, sslverify=True, timeout=8):
    """
    Resolve the address of an ENS domain
    """

    url = whoisBaseUrl.join(URL('resolve/address/{domain}'.format(
        domain=domain)))

    try:
        with async_timeout.timeout(timeout):
            async with aiohttp.ClientSession() as session:
                async with session.get(str(url), verify_ssl=sslverify) as resp:
                    data = await resp.json()
                    return data['result']['result']
    except asyncio.TimeoutError:
        return None
    except Exception:
        return None



import os.path
import asyncio

from PyQt5.QtWidgets import QMenu
from PyQt5.QtWidgets import QHeaderView
from PyQt5.QtWidgets import QComboBox
from PyQt5.QtWidgets import QWidget

from PyQt5.QtGui import QStandardItemModel
from PyQt5.QtCore import QObject
from PyQt5.QtCore import QMimeData
from PyQt5.QtCore import QCoreApplication
from PyQt5.QtCore import QUrl
from PyQt5.QtWidgets import QAbstractItemView

from galacteek import ensure, log
from galacteek.ipfs.wrappers import ipfsOp
from galacteek.ipfs.cidhelpers import *
from galacteek.ipfs.ipfsops import *
from galacteek.core.modelhelpers import *
from galacteek.dweb.atom import DWEB_ATOM_FEEDFN

from . import ui_peersmgr
from .widgets import *
from .helpers import *
from .dialogs import *


def iUsername():
    return QCoreApplication.translate('PeersManager', 'Username')


def iPeerId():
    return QCoreApplication.translate('PeersManager', 'Peer ID')


def iPingAvg():
    return QCoreApplication.translate('PeersManager', 'Ping Average (ms)')


def iLocation():
    return QCoreApplication.translate('PeersManager', 'Location')


def iInvalidDagCID(dagCid):
    return QCoreApplication.translate('PeersManager',
                                      'Invalid DAG CID: {0}').format(dagCid)


class PeersModel(QStandardItemModel):
    def mimeData(self, indexes):
        mimedata = QMimeData()

        for idx in indexes:
            idxPeer = self.index(idx.row(), 1, idx.parent())
            if idxPeer.isValid():
                peer = self.data(idxPeer)
                mimedata.setUrls([QUrl('galacteekpeer:{}'.format(peer))])
                break

        return mimedata

    def canDropMimeData(self, data, action, row, column, parent):
        return True

    def dropMimeData(self, data, action, row, column, parent):
        if data.hasUrls():
            return True


class PeersTracker(QObject):
    def __init__(self, ctx):
        super().__init__()
        self.model = PeersModel()
        self.model.setHorizontalHeaderLabels([
            iUsername(),
            iPeerId(),
            iPingAvg(),
            iLocation(),
            ''
        ])
        self.ctx = ctx
        self.ctx.peers.changed.connect(self.onPeersChange)
        self.ctx.peers.peerAdded.connect(self.onPeerAdded)
        self.ctx.peers.peerModified.connect(self.onPeerModified)
        self.ctx.peers.peerLogout.connect(self.onPeerLogout)
        self._peersRows = {}

    @property
    def modelRoot(self):
        return self.model.invisibleRootItem()

    @property
    def peersRows(self):
        return self._peersRows

    def onPeerLogout(self, peerId):
        modelSearch(self.model,
                    parent=self.modelRoot.index(),
                    search=peerId, delete=True)
        if peerId in self._peersRows:
            del self._peersRows[peerId]

    def onPeerModified(self, peerId):
        peerCtx = self.ctx.peers.getByPeerId(peerId)
        if not peerCtx:
            return

        row = self._peersRows.get(peerId, None)
        if row:
            row[0].setText(peerCtx.ident.username)
            row[1].setText(peerId)
            row[3].setText(peerCtx.ident.location)

    def onPeerAdded(self, peerId):
        peerCtx = self.ctx.peers.getByPeerId(peerId)
        if not peerCtx:
            return

        self.addPeerFromEntry(peerId, peerCtx)

    def addPeerFromEntry(self, peerId, peerCtx):
        row = [
            UneditableItem(peerCtx.ident.username),
            UneditableItem(peerId),
            UneditableItem(str(peerCtx.pingavg)),
            UneditableItem(peerCtx.ident.location),
            UneditableItem(''),
        ]
        rowStatus = [
            UneditableItem('Joined: {0}'.format(peerCtx.ident.dateCreated)),
            UneditableItem(peerId),
        ]

        self.modelRoot.appendRow(row)
        row[0].appendRow(rowStatus)
        self._peersRows[peerId] = row

    def onPeersChange(self):
        pass


class PeersManager(GalacteekTab):
    def __init__(self, gWindow, peersTracker, **kw):
        super().__init__(gWindow, **kw)

        self.lock = asyncio.Lock()
        self.peersTracker = peersTracker

        self.peersWidget = QWidget()
        self.addToLayout(self.peersWidget)
        self.ui = ui_peersmgr.Ui_PeersManager()
        self.ui.setupUi(self.peersWidget)

        self.ui.tree.setModel(self.peersTracker.model)
        self.ui.tree.header().setSectionResizeMode(
            0, QHeaderView.ResizeToContents)
        self.ui.tree.header().setSectionResizeMode(
            1, QHeaderView.ResizeToContents)
        self.ui.tree.header().setSectionResizeMode(
            2, QHeaderView.ResizeToContents)

        self.ui.tree.setContextMenuPolicy(Qt.CustomContextMenu)
        self.ui.tree.customContextMenuRequested.connect(self.onContextMenu)
        self.ui.tree.doubleClicked.connect(self.onDoubleClick)
        self.ui.tree.setDragDropMode(QAbstractItemView.DragOnly)

        self.ui.search.returnPressed.connect(self.onSearch)

        self.app.ipfsCtx.peers.changed.connect(self.onPeersChange)
        self.app.ipfsCtx.peers.peerAdded.connect(self.onPeerAdded)
        ensure(self.refreshControls())

    @property
    def model(self):
        return self.peersTracker.model

    def onSearch(self):
        search = self.ui.search.text()
        if len(search) > 0:
            self.ui.tree.keyboardSearch(search)

    def onPeersChange(self):
        self.ui.peersCountLabel.setText(str(self.app.ipfsCtx.peers.peersCount))

    def onPeerAdded(self, peerId):
        ensure(self.refreshControls())

    async def refreshControls(self):
        def onPeerAction(idx, id):
            peerCtx = self.peersTracker.ctx.peers.getByPeerId(id)
            if idx in [0, 1]:
                method = 'ipns' if idx == 1 else 'direct'
                ensure(self.explorePeerHome(peerCtx, method=method))
            elif idx == 2:
                ensure(self.followPeerFeed(peerCtx))

        def followPeer(id):
            peerCtx = self.peersTracker.ctx.peers.getByPeerId(id)

            runDialog(AddFeedDialog, self.app.marksLocal,
                      joinIpns(peerCtx.ident.dagIpns),
                      feedName=peerCtx.ident.username)

        with await self.lock:
            for peerId, row in self.peersTracker.peersRows.items():
                idx = self.model.indexFromItem(row[4])
                if self.ui.tree.indexWidget(idx) is not None:
                    continue

                btnHomeCombo = QComboBox()
                btnHomeCombo.addItem('Browse homepage (direct)')
                btnHomeCombo.addItem('Browse homepage (IPNS)')
                btnHomeCombo.addItem('Follow (Atom feed)')
                btnHomeCombo.activated.connect(
                    lambda idx: onPeerAction(idx, peerId))

                self.ui.tree.setIndexWidget(idx, btnHomeCombo)

    def onContextMenu(self, point):
        idx = self.ui.tree.indexAt(point)
        if not idx.isValid():
            return

        menu = QMenu(self)
        menu.exec(self.ui.tree.mapToGlobal(point))

    def onDoubleClick(self, idx):
        idxPeerId = self.model.sibling(idx.row(), 1, idx)
        peerId = self.model.data(idxPeerId)
        peerCtx = self.peersTracker.ctx.peers.getByPeerId(peerId)

        ensure(self.explorePeerHome(peerCtx))

    @ipfsOp
    async def followPeerFeed(self, op, peerCtx):
        identMsg = peerCtx.ident
        path = os.path.join(joinIpns(identMsg.dagIpns), DWEB_ATOM_FEEDFN)
        ipfsPath = IPFSPath(path)

        try:
            await self.app.sqliteDb.feeds.follow(ipfsPath.dwebUrl)
        except Exception:
            # TODO
            pass

    @ipfsOp
    async def explorePeerHome(self, op, peerCtx, method='direct'):
        identMsg = peerCtx.ident

        dagCid = identMsg.dagCid
        if not cidValid(dagCid):
            log.debug('invalid DAG CID: {}'.format(dagCid))
            return messageBox(iInvalidDagCID(dagCid))

        if method == 'ipns':
            self.gWindow.addBrowserTab().browseFsPath(
                os.path.join(joinIpns(identMsg.dagIpns), 'index.html'))
        elif method == 'direct':
            self.gWindow.addBrowserTab().browseFsPath(
                os.path.join(joinIpfs(identMsg.dagCid), 'index.html'))



import asyncio
import aiosqlite
import sqlite3
import time
import weakref
from datetime import datetime

from PyQt5.QtCore import QUrl
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal

from galacteek import log
from galacteek import ensure
from galacteek.ipfs import ipfsOp
from galacteek.core.schemes import SCHEME_DWEB
from galacteek.core.schemes import SCHEME_IPFS
from galacteek.dweb.atom import IPFSAtomFeedParser
from galacteek.dweb.atom import IPFSAtomFeedEntry
from galacteek.dweb.atom import IPFSAtomFeed
from galacteek.dweb.atom import AtomFeedExistsError
from galacteek.dweb.atom import AtomParseFeedError
from galacteek.ipfs.cidhelpers import IPFSPath


schemaScript = '''
CREATE TABLE if not exists url_history_items
(id integer primary key, url text, scheme text,
rootcid text, rootcidv integer, ctime timestamp);

CREATE TABLE if not exists url_history_visits
(history_item_id integer, title text, atime timestamp);

CREATE TABLE if not exists atom_feeds
(id integer primary key, url text, scheme text, feed_id text,
category varchar(1, 128), autopin_entries integer, ctime timestamp);

CREATE TABLE if not exists atom_feed_history
(id integer primary key, atom_feed_id integer,
objpath text, status integer);

CREATE TABLE if not exists atom_feed_entries
(id integer primary key, atom_feed_id integer, entry_id text,
status integer, published timestamp);
'''


class SqliteDatabase:
    def __init__(self, dbPath):
        self._path = dbPath
        self._db = None
        self.feeds = AtomFeedsDatabase(self)

    @property
    def db(self):
        return self._db

    async def setup(self):
        try:
            self._db = await aiosqlite.connect(self._path)
            self._db.row_factory = aiosqlite.Row
        except:
            return False

        def adapt_datetime(ts):
            return time.mktime(ts.timetuple())

        sqlite3.register_adapter(datetime, adapt_datetime)

        try:
            await self.db.executescript(schemaScript)
        except Exception:
            log.debug('Error while executing schema script')
            return False

    async def historySearch(self, input):
        rows = []
        query = '''
        SELECT DISTINCT url, title, scheme, rootcid, rootcidv, id
        FROM url_history_visits
        INNER JOIN url_history_items ON
            url_history_items.id = url_history_visits.history_item_id
        WHERE url LIKE '%{input}%' OR title LIKE '%{input}%'
        COLLATE NOCASE
        ORDER BY atime desc'''

        _urls = []
        async with self.db.execute(query.format(input=input)) as cursor:
            async for row in cursor:
                if row['url'] in _urls:
                    continue

                _urls.append(row['url'])

                await asyncio.sleep(0)
                rows.append(row)
        return rows

    async def historyRecord(self, url, title):
        urlItemId = None
        existing = await self.historySearch(url)

        if len(existing) > 0:
            urlItemId = existing.pop()['id']
        else:
            qUrl = QUrl(url)
            scheme = qUrl.scheme() if qUrl.isValid() else ''

            rootcid = ''
            rootcidv = None

            if scheme in [SCHEME_DWEB, SCHEME_IPFS]:
                ipfsPath = IPFSPath(url)
                if ipfsPath.valid and ipfsPath.rootCid:
                    rootcid = str(ipfsPath.rootCid)
                    rootcidv = ipfsPath.rootCid.version

            cursor = await self.db.execute(
                """INSERT INTO url_history_items
                (url, scheme, rootcid, rootcidv, ctime)
                VALUES (?, ?, ?, ?, ?)""",
                (url, scheme, rootcid, rootcidv, datetime.now()))
            urlItemId = cursor.lastrowid

        cursor = await self.db.execute(
            """INSERT INTO url_history_visits
            (history_item_id, title, atime) VALUES (?, ?, ?)""",
            (urlItemId, title, datetime.now()))

        await self.db.commit()
        return True

    async def historyClear(self, before=None):
        if before:
            cursor = await self.db.execute(
                "DELETE FROM url_history_items WHERE ctime < {0}".format(
                    before))
            log.debug('Cleared {} history records'.format(cursor.rowcount))

            await self.db.execute(
                "DELETE FROM url_history_visits WHERE atime < {0}".format(
                    before))
        else:
            # Clear all
            cursor = await self.db.execute(
                'DELETE FROM url_history_items'
            )
            log.debug('Cleared {} history records'.format(cursor.rowcount))
            await self.db.execute(
                'DELETE FROM url_history_visits'
            )

        await self.db.commit()

    async def close(self):
        await self.db.close()


class AtomFeedsDatabase(QObject):
    processedFeedEntry = pyqtSignal(IPFSAtomFeed, IPFSAtomFeedEntry)
    feedRemoved = pyqtSignal(str)

    def __init__(self, database, parent=None):
        super(AtomFeedsDatabase, self).__init__(parent)

        self.loop = asyncio.get_event_loop()
        self.parser = IPFSAtomFeedParser()
        self.sqliteDb = database
        self.lock = asyncio.Lock()

        self._handled_by_id = weakref.WeakValueDictionary()
        self._process_task = None

    @property
    def db(self):
        return self.sqliteDb.db

    def feedFromId(self, feedId):
        return self._handled_by_id.get(feedId)

    async def unfollow(self, feedId):
        with await self.lock:
            feed = await self.getFromId(feedId)

            if feed:
                if feedId in self._handled_by_id:
                    del self._handled_by_id[feedId]

                params = {'id': feed['id']}
                query = '''
                    DELETE FROM atom_feed_entries
                    WHERE atom_feed_id=:id
                '''
                await self.db.execute(query, params)

                query = '''
                    DELETE FROM atom_feed_history
                    WHERE atom_feed_id=:id
                '''
                await self.db.execute(query, params)

                query = '''
                    DELETE FROM atom_feeds
                    WHERE id=:id
                '''
                await self.db.execute(query, params)
                await self.db.commit()

                self.feedRemoved.emit(feedId)
                return True

    @ipfsOp
    async def follow(self, ipfsop, url):
        if not isinstance(url, str):
            raise ValueError('Wrong URL')

        path = IPFSPath(url)
        if not path.valid:
            raise ValueError('Wrong URL')

        scheme = 'dweb'

        try:
            feed = await self.parser.parse(path)
        except AtomParseFeedError:
            return False

        resolved = await ipfsop.resolve(str(path))
        if not resolved:
            raise ValueError('Does not resolve')

        # Pin the resolved object (yeah!)
        await ipfsop.pin(resolved)

        #
        # Here's why Atom's a great choice over RSS for the dweb
        # Atom feeds *must* have an identifier
        # Using an IPNS path as the Atom feed identifier = uniqueness
        #

        # Same ID
        if await self.getFromId(feed.id):
            raise AtomFeedExistsError()

        # Same ID and URL (not necessarily the same)
        if await self.getFromUrlAndId(url, feed.id):
            raise AtomFeedExistsError()

        await self.db.execute(
            """INSERT INTO atom_feeds
            (url, scheme, feed_id, autopin_entries, ctime)
            VALUES (?, ?, ?, ?, ?)""",
            (url, scheme, feed.id, 1, datetime.now()))

        await self.db.commit()
        sqlObj = await self.getFromUrlAndId(url, feed.id)

        await self.db.execute(
            """INSERT INTO atom_feed_history
            (atom_feed_id, objpath, status)
            VALUES (?, ?, ?)""",
            (sqlObj['id'], resolved, 0))

        await self.processFeed(sqlObj)
        return True

    async def getFromId(self, feedId):
        query = '''
        SELECT *
        FROM atom_feeds
        WHERE feed_id=:feedid
        LIMIT 1
        '''
        cursor = await self.db.execute(query, {'feedid': feedId})
        return await cursor.fetchone()

    async def getFromUrlAndId(self, url, feedId):
        query = '''
        SELECT *
        FROM atom_feeds
        WHERE feed_id=:feedid AND url=:url
        LIMIT 1
        '''
        cursor = await self.db.execute(query,
                                       {'feedid': feedId, 'url': url})
        return await cursor.fetchone()

    async def allFeeds(self):
        query = 'SELECT * FROM atom_feeds'
        return await self.rows(query)

    @ipfsOp
    async def processFeed(self, ipfsop, feedSql):
        with await self.lock:
            path = IPFSPath(feedSql['url'])
            resolved = await ipfsop.resolve(path.objPath)

            if not resolved:
                # TODO
                return

            await ipfsop.sleep()

            needParse = True
            historyObj = await self.feedObjectHistory(
                feedSql['id'], resolved)

            if not historyObj:
                historyObjId = await self.feedNewObjectHistory(
                    feedSql['id'], resolved)
            else:
                historyObjId = historyObj['id']
                needParse = False

            atomFeed = self.feedFromId(feedSql['feed_id'])

            if not atomFeed or needParse:
                try:
                    atomFeed = await self.parser.parse(resolved)
                except AtomParseFeedError:
                    log.debug('Atom: failed to parse {p}'.format(
                        p=resolved))
                    return False

            self._handled_by_id[atomFeed.id] = atomFeed

            entries = await self.searchEntries(atomFeed.id)
            existingIds = [ent['entry_id'] for ent in entries]

            for entry in atomFeed.entries:
                await asyncio.sleep(0.2)
                if not isinstance(entry.id, str):
                    continue

                if entry.id not in existingIds:
                    _rid = await self.addEntry(feedSql['id'], entry.id)

                    entry.status = IPFSAtomFeedEntry.ENTRY_STATUS_NEW
                    entry.srow_id = _rid
                    self.processedFeedEntry.emit(atomFeed, entry)

                    if feedSql['autopin_entries'] == 1 and \
                            feedSql['scheme'] == 'dweb':
                        path = IPFSPath(entry.id)
                        if path.valid:
                            log.debug('Atom: autopinning {id}'.format(
                                id=entry.id))
                            ensure(ipfsop.pin(path.objPath))
                else:
                    for exent in entries:
                        if exent['entry_id'] == entry.id:
                            entry.status = exent['status']
                            entry.srow_id = exent['id']
                            self.processedFeedEntry.emit(atomFeed,
                                                         entry)

            # Mark the object as processed
            await self.feedObjectHistoryUpdateStatus(historyObjId,
                                                     resolved, 1)

    async def searchEntries(self, feedId):
        query = '''
        SELECT atom_feed_entries.id,
               atom_feed_entries.entry_id,
               atom_feed_entries.status
        FROM atom_feed_entries
        INNER JOIN atom_feeds ON
            atom_feeds.id = atom_feed_entries.atom_feed_id
        WHERE atom_feeds.feed_id=:feedid
        '''

        return await self.rows(query, {'feedid': feedId})

    async def feedHasObjectInHistory(self, feedId, objPath, status=0):
        query = '''
        SELECT DISTINCT atom_feed_history.objpath
        FROM atom_feed_history
        INNER JOIN atom_feeds ON
            atom_feeds.id = atom_feed_history.atom_feed_id
        WHERE atom_feeds.feed_id=:feedid AND status=:status AND
            atom_feed_history.objpath=:objpath
        '''

        params = {'feedid': feedId, 'status': status, 'objpath': objPath}
        cursor = await self.db.execute(query, params)
        return await cursor.fetchone()

    async def feedObjectHistory(self, feedId, objPath):
        query = '''
        SELECT *
        FROM atom_feed_history
        INNER JOIN atom_feeds ON
            atom_feeds.id = atom_feed_history.atom_feed_id
        WHERE atom_feeds.id=:feedid AND
            atom_feed_history.objpath=:objpath
        '''

        params = {'feedid': feedId, 'objpath': objPath}
        cursor = await self.db.execute(query, params)
        return await cursor.fetchone()

    async def feedObjectHistoryUpdateStatus(self, historyId, objPath, status):
        query = '''
        UPDATE atom_feed_history
        SET status=:status
        WHERE id=:historyid AND objpath=:objpath
        '''

        params = {'historyid': historyId, 'objpath': objPath, 'status': status}
        await self.db.execute(query, params)
        await self.db.commit()

    async def feedNewObjectHistory(self, feedId, objPath):
        cursor = await self.db.execute(
            """INSERT INTO atom_feed_history
            (atom_feed_id, objpath, status)
            VALUES (?, ?, ?)""",
            (feedId, objPath, 0))
        await self.db.commit()
        return cursor.lastrowid

    async def rows(self, query, params={}):
        rows = []
        async with self.db.execute(query, params) as cursor:
            async for row in cursor:
                await asyncio.sleep(0)
                rows.append(row)
        return rows

    async def addEntry(self, feedId, entryId,
                       status=IPFSAtomFeedEntry.ENTRY_STATUS_NEW):
        cursor = await self.db.execute(
            """INSERT INTO atom_feed_entries
            (atom_feed_id, entry_id, status, published)
            VALUES (?, ?, ?, ?)""",
            (feedId, entryId, status, datetime.now()))
        await self.db.commit()
        return cursor.lastrowid

    async def feedEntrySetStatus(self, entryId, status):
        query = '''
        UPDATE atom_feed_entries
        SET status=:status
        WHERE id=:id
        '''

        params = {'id': entryId, 'status': status}
        await self.db.execute(query, params)
        await self.db.commit()

    async def start(self):
        if not self._process_task:
            self._process_task = self.loop.create_task(self.processTask())

    async def processTask(self):
        while True:
            feeds = await self.allFeeds()

            for feed in feeds:
                await self.processFeed(feed)

            await asyncio.sleep(60)
